==> r9p-0.3.tgz: r9p-0.3/r9p/client.rb <==
# Copyright (C) 2007 Kris Maglione
# See PERMISSIONS

module R9P
	class Exception <Exception
	end
	class ProtocolException <Exception
	end
	class Client
		RootFid = 0
		module Constants
			OREAD = 0x00
			OWRITE = 0x01
			ORDWR = 0x02
			OEXEC = 0x03
			OEXCL = 0x04
			OTRUNC = 0x10
			OREXEC = 0x20
			ORCLOSE = 0x40
			OAPPEND = 0x80
		end
		include Constants
		include Qlock
		attr_reader :msize

		def initialize(con, root = nil)
			qlinit
			@lastfid = RootFid
			@fids = []
			@files = {}
			@mux = Mux.new(con, :maxtag => 256) { |dat|
				fcall = Fcall.from dat
			}

			resp = dorpc Fcall::Tversion, :version => P9Version, :msize => 65535
			if resp.version != P9Version
				throw ProtocolException.new("Can't speak 9P version '#{resp.version}'")
			end
			@msize = resp.msize

			dorpc Fcall::Tattach, :fid => RootFid, :afid => NoFid, :uname => ENV['USER'], :aname => ''

			if root
				path = root.split('/').reject{|v| v=='' }
				resp = dorpc Fcall::Twalk, :fid => RootFid, :newfid => RootFid, :wname => path
			end

			if block_given?
				begin
					yield self
				ensure
					cleanup
				end
			end
			return
		ensure
			@mux.fd.close if @mux
		end

		def cleanup
			@files.each_value { |f|
				f.close
			}
		ensure
			@mux.fd.close
			@mux = nil
		end

		def dorpc(clas, *args)
			fcall = clas.new *args
			resp = @mux.rpc fcall
			if resp.kind_of? Fcall::Rerror
				throw Exception.new("RPC returned error: #{fcall.class.inspect} #{fcall.fields.inspect}: #{resp.ename}")
			end
			if fcall.type != resp.type ^ 1
				throw ProtocolException.new("Missmatched RPC message types: #{fcall.class.inspect} => #{resp.class.inspect}")
			end
			resp
		end

		def getfid
			qsync {
				@fids.pop || @lastfid += 1
			}
		end
		def putfid(fid)
			qsync {
				if @files.delete(fid)
					@fids.push fid
				else
					throw Exception.new("Attempt to free unused fid")
				end
			}
		end

		def open(path, mode = OREAD)
			path = path.split('/').reject{|v| v=='' }
			fid = getfid

			resp = dorpc Fcall::Twalk, :fid => RootFid, :newfid => fid, :wname => path
			file = File.new self, fid, mode
			@files[fid] = file

			if block_given?
				begin
					yield file
				ensure
					file.close
				end
			else
				file
			end
		end

		class File
			include Stat::Constants
			include Client::Constants
			include Qlock
			def initialize(client, fid, mode)
				qlinit
				@fid = fid
				@client = client
				@offset = 0
				@mode = mode
				@fd = nil

				fcall = dorpc Fcall::Topen, :mode => mode
				@iounit = fcall.iounit
				@qid = fcall.qid
			end

			def stat
				resp = dorpc Fcall::Tstat
				resp.stat
			end

			def read(count = @iounit, offset = @offset, buf = '')
				qsync {
					while count > 0
						n = [count, @iounit].min
						count -= n

						resp = dorpc Fcall::Tread, :offset => @offset, :count => n
						data = resp.data

						@offset += data.length
						buf << data

						if data.length < n
							break
						end
					end
				}
				if buf.length == 0
					nil
				else
					buf
				end
			end
			def write(data, offset = @offset)
				off = 0
				qsync {
					while off < data.length
						n = [data.length, @iounit].min

						resp = dorpc Fcall::Twrite, :offset => @offset, :data => data[off..off+n-1]

						off += resp.count
						@offset += resp.count
						if resp.count < n
							break
						end
					end
				}
				off
			end
			def readdir
				unless @qid.type.to_i&Qid::QTDIR
					raise Exception.new("Can only call readdir on a directory")
				end
				data = ''
				begin
					dat = read(@iounit, data.length)
					data << dat if dat
				end while dat
				Stat.from_data data
			end

			def close
				fd_close if @fd
			ensure
				qsync {
					dorpc Fcall::Tclunk
				}
				@client.putfid @fid

				@tg = @fid = @client = @qid = nil
			end
			def fd_close
				@fd.close
			ensure
				@fd = nil
				@tg.list.each { |thr|
					thr.kill if thr.alive?
					thr.join
				}
			end

			def to_fd
				if @fd
					throw Exception.new("File descriptor already open for file")
				end

				fds = Socket::pair Socket::PF_UNIX, Socket::SOCK_STREAM, 0
				@tg = ThreadGroup.new
				mtx = Mutex.new

				reader = nil
				writer = nil
				mtx.lock
				@tg.add reader = Thread.new(fds[1]) { |fd|
					mtx.synchronize {
						unless (@mode&3) == OWRITE ||  (@mode&3) == ORDWR
							fd.close unless writer.alive?
							return
						end
					}
					while data = fd.sysread(@iounit)
						if write(data) == 0
							break
						end
					end
					mtx.synchronize {
						if fd.eof
							writer.terminate
							fd.close
						elsif !writer.alive?
							fd.close
						end
					}
				}
				@tg.add writer = Thread.new(fds[1]) { |fd|
					mtx.synchronize {
						unless (@mode&3) == OREAD ||  (@mode&3) == ORDWR
							fd.close unless reader.alive
							return
						end
					}
					len = 1
					while data = read
						while data.length > 0
							len = fd.write data
							if len == 0
								break
							end
							data.slice!(0..len-1)
						end
						if len == 0
							break
						end
					end
					mtx.synchronize {
						reader.terminate
						fd.close
					}
				}
				mtx.unlock

				@fd = fds[0]
				if block_given?
					begin
						yield @fd
					ensure
						fd_close
					end
				else
					@fd
				end
			end

			def dorpc(clas, args = {})
				@client.dorpc clas, args.merge!(:fid => @fid)
			end
			private :dorpc
		end
	end
end

==> r9p-0.4.tgz: r9p-0.4/r9p/client.rb <==
# Copyright (C) 2007 Kris Maglione
# See PERMISSIONS

module R9P
	class Exception <Exception
	end
	class ProtocolException <R9P::Exception
	end
	class RPCError <R9P::Exception
	end
	class Client
		RootFid = 0
		module Constants
			OREAD = 0x00
			OWRITE = 0x01
			ORDWR = 0x02
			OEXEC = 0x03
			OEXCL = 0x04
			OTRUNC = 0x10
			OREXEC = 0x20
			ORCLOSE = 0x40
			OAPPEND = 0x80
		end
		include Constants
		include Qlock

		def initialize(con, root = nil)
			qlinit
			@lastfid = RootFid
			@fids = []
			@files = {}
			@mux = Mux.new(con, :maxtag => 256) { |dat|
				fcall = Fcall.from dat
			}

			resp = dorpc Fcall::Tversion, :version => R9P::Version, :msize => 65535
			if resp.version != R9P::Version
				raise ProtocolException, "Can't speak 9P version '#{resp.version}'"
			end
			@msize = resp.msize

			dorpc Fcall::Tattach, :fid => RootFid, :afid => Fcall::NoFid, :uname => ENV['USER'], :aname => ''

			if root
				path = splitpath(root)
				resp = dorpc Fcall::Twalk, :fid => RootFid, :newfid => RootFid, :wname => path
			end

			if block_given?
				begin
					yield self
				ensure
					cleanup
				end
			end
			return
		ensure
			@mux.fd.close if @mux
		end

		def cleanup
			@files.each_value { |f|
				f.close
			}
		ensure
			@mux.fd.close
			@mux = nil
		end

		def dorpc(clas, *args)
			fcall = clas.new *args
			resp = @mux.rpc fcall
			if resp.kind_of? Fcall::Rerror
				raise RPCError, "RPC returned error (#{fcall.class.inspect}: #{fcall.fields.inspect}): #{resp.ename}"
			end
			if fcall.type != resp.type ^ 1
				raise ProtocolException, "Missmatched RPC message types: #{fcall.class.inspect} => #{resp.class.inspect}"
			end
			resp
		end

		def splitpath(path)
			path.split('/').reject{|v| v=='' }
		end
		private:splitpath

		def getfid
			qsync {
				@fids.pop || @lastfid += 1
			}
		end
		def putfid(fid)
			qsync {
				@files.delete(fid)
				@fids.push fid
			}
		end
		private :getfid, :putfid

		def clunk(fid)
			dorpc Fcall::Tclunk, :fid => fid
		ensure
			putfid fid
		end
		private :clunk

		def walk(path)
			path = path.clone
			fid = getfid
			ofid = RootFid
			begin
				dorpc Fcall::Twalk, :fid => ofid, :newfid => fid, :wname => path.slice!(0..Fcall::MaxWelem-1)
				ofid = fid
			end while path.length > 0
			begin
				yield fid
			rescue
				clunk fid
				raise
			end
		end

		def _open(path, mode, block)
			resp = fid = file = nil

			walk(path) do |fid|
				fid = fid
				resp = yield(fid)
			end

			file = File.new(self, resp, fid, mode) do
				clunk fid
			end
			@files[fid] = file

			if block
				begin
					block.call file
				ensure
					file.close
				end
			else
				file
			end
		end
		private :_open, :walk

		def open(path, mode = OREAD, &block)
			path = splitpath(path)

			_open(path, mode, block) do |fid|
				dorpc Fcall::Topen, :fid => fid, :mode => mode
			end
		end

		def create(path, mode = OREAD, perm = 0, &block)
			path = splitpath(path)
			name = path.pop

			_open(path, mode, block) do |fid|
				dorpc Fcall::Tcreate, :fid => fid, :mode => mode, :name => name, :perm => perm
			end
		end

		def remove(path)
			path = splitpath(path)

			walk(path) do |fid|
				dorpc Fcall::Tremove, :fid => fid
			end
			true
		end

		def stat(path)
			path = splitpath(path)

			walk(path) do |fid|
				begin
					resp = dorpc Fcall::Tstat, :fid => fid
					resp.stat
				ensure
					clunk fid
				end
			end
		rescue RPCError
			return nil
		end

		class File
			include Stat::Constants
			include Client::Constants
			include Qlock
			attr_reader :fid, :qid, :mode, :offset, :iounit
			def initialize(client, fcall, fid, mode, &cleanup)
				qlinit
				@client = client
				@fid = fid
				@cleanup = cleanup
				@mode = mode
				@iounit = fcall.iounit
				@qid = fcall.qid

				@offset = 0
				@fd = nil
			end

			def dorpc(clas, args = {})
				@client.dorpc clas, args.merge!(:fid => @fid)
			end
			private :dorpc

			def stat
				resp = dorpc Fcall::Tstat
				resp.stat
			end

			def read(count = @iounit, offset = @offset, buf = '')
				qsync {
					while count > 0
						n = [count, @iounit].min
						count -= n

						resp = dorpc Fcall::Tread, :offset => @offset, :count => n
						data = resp.data

						@offset += data.length
						buf << data

						if data.length < n
							break
						end
					end
				}
				if buf.length == 0
					nil
				else
					buf
				end
			end
			def write(data, offset = @offset)
				off = 0
				qsync {
					while off < data.length
						n = [data.length, @iounit].min

						resp = dorpc Fcall::Twrite, :offset => @offset, :data => data[off..off+n-1]

						off += resp.count
						@offset += resp.count
						if resp.count < n
							break
						end
					end
				}
				off
			end
			def readdir
				unless @qid.type.to_i&Qid::QTDIR
					raise Exception, "Can only call readdir on a directory"
				end
				data = ''
				begin
					dat = read(@iounit, data.length)
					data << dat if dat
				end while dat
				Stat.from_data data
			end

			def close
				fd_close if @fd
			ensure
				@cleanup.call self

				@tg = @fid = @client = @qid = nil
			end
			def fd_close
				@fd.close
			ensure
				@fd = nil
				@tg.list.each { |thr|
					thr.kill if thr.alive?
					thr.join
				}
			end

			def to_fd
				if @fd
					raise Exception, "File descriptor already open for file"
				end

				fds = Socket::pair Socket::PF_UNIX, Socket::SOCK_STREAM, 0
				@tg = ThreadGroup.new
				mtx = Mutex.new

				reader = nil
				writer = nil
				mtx.lock
				@tg.add reader = Thread.new(fds[1]) { |fd|
					mtx.synchronize {
						unless (@mode&3) == OWRITE ||  (@mode&3) == ORDWR
							fd.close unless writer.alive?
							return
						end
					}
					while data = fd.sysread(@iounit)
						if write(data) == 0
							break
						end
					end
					mtx.synchronize {
						if fd.eof
							writer.terminate
							fd.close
						elsif !writer.alive?
							fd.close
						end
					}
				}
				@tg.add writer = Thread.new(fds[1]) { |fd|
					mtx.synchronize {
						unless (@mode&3) == OREAD ||  (@mode&3) == ORDWR
							fd.close unless reader.alive
							return
						end
					}
					len = 1
					while data = read
						while data.length > 0
							len = fd.write data
							if len == 0
								break
							end
							data.slice!(0..len-1)
						end
						if len == 0
							break
						end
					end
					mtx.synchronize {
						reader.terminate
						fd.close
					}
				}
				mtx.unlock

				@fd = fds[0]
				if block_given?
					begin
						yield @fd
					ensure
						fd_close
					end
				else
					@fd
				end
			end
		end
	end
end
